#ifndef MAZE_H
#define MAZE_H

#include <stdlib.h>

struct vertex
{
	unsigned long long int id;

	// Pointers to the neighbours of the vertex.
	struct vertex *top;
	struct vertex *left;
	struct vertex *bottom;
	struct vertex *right;

	// Pointer to the parent vertex in the BFS algorithm.
	struct vertex *parent;

	// Boolean to know if the vertex is the exit of the graph.
	unsigned int is_end;
};

struct graph
{
	struct vertex *vertices;

	unsigned long long int width;
	unsigned long long int height;
};

/*
 * Preconditions: - step must be a positive integer.
 * Postconditions: None.
 *
 * Returns an integer n equals to min + k * step (k >= 0) such that
 * n is between min and max included.
 */
long long int random_int(long long int min, long long int max, int const step);

/*
 * Preconditions: None.
 * Postconditions: None.
 *
 * Returns a pointer on the graph, or NULL on error.
 */
struct graph *init_graph(long long int const width, long long int const height);

/*
 * Preconditions: None.
 * Postconditions: If a wall has been added, graph has lost an edge.
 *                 Otherwise, graph is not modified, or NULL.
 */
void add_wall(struct graph *const graph, long long int const i, long long int const j);

/*
 * Preconditions: - graph must be initialized.
 *                - first_vertex_id must be the first vertex id of graph.
 *                - cur_width must be equal to the width of graph.
 *                - cur_height must be equal to the height of graph.
 * Postconditions: graph represents a maze randomly generated.
 */
void generate_maze_aux(struct graph *const graph, long long int const first_vertex_id,
	long long int const cur_width, long long int const cur_height);

/*
 * Preconditions: None.
 * Postconditions: graph represents a maze randomly generated, or NULL.
 */
void generate_maze(struct graph *const graph);

/*
 * Preconditions: None.
 * Postconditions: The memory allocated for graph by init_graph() is freed,
 *                 or does nothing if graph is NULL.
 */
void free_graph(struct graph *graph);

/*
 * Preconditions: None.
 * Postconditions: None.
 *
 * Returns a NULL-terminated list of the vertex's neighbours.
 */
struct vertex **get_neighbours(struct vertex *const vertex);

/*
 * Preconditions: If a call to find_exit() has already happen for a graph,
 *				  the said graph must reset all vertices parent to NULL.
 * Postconditions: Some vertices (or all) will have a new parent vertex.
 *
 * Returns the exit vertex which contains the path to the start vertex,
 * or NULL on error.
 */
struct vertex *find_exit(struct graph *const graph, long long int const start);

#endif //MAZE_H
