#include "maze.h"

long long int random_int(long long int min, long long int max, int const step)
{
	if (min > max)
	{
		int tmp = min;
		min = max;
		max = tmp;
	}

	long long int k = (max - min) / step;

	return min + step * (rand() % (k+1));
}

struct graph *init_graph(long long int const width, long long int const height)
{
	if (width < 1 || height < 1)
		return NULL;

	long long int const size = width * height;

	struct graph *graph = (struct graph *) calloc(1, sizeof(struct graph));
	if (graph == NULL)
		return NULL;

	graph->vertices = (struct vertex *) calloc(size, sizeof(struct vertex));
	if (graph->vertices == NULL)
	{
		free(graph);
		return NULL;
	}

	graph->width = width;
	graph->height = height;

	// If the maze is only one cell
	if (width == 1 && height == 1)
	{
		graph->vertices[0].id = 0;
		graph->vertices[0].parent = NULL;
		graph->vertices[0].is_end = 1;

		graph->vertices[0].top = NULL;
		graph->vertices[0].bottom = NULL;
		graph->vertices[0].left = NULL;
		graph->vertices[0].right = NULL;

		return graph;
	}

	for (long long int i = 0; i < size; i++)
	{
		graph->vertices[i].id = i;
		graph->vertices[i].parent = NULL;
		graph->vertices[i].is_end = (i == size-1) ? 1 : 0;

		if (i < width) // If it is a top cell of the maze
		{
			graph->vertices[i].top = NULL;
			graph->vertices[i].bottom = (height == 1) ? NULL : &graph->vertices[i+width];
			graph->vertices[i].left = (i == 0) ? NULL : &graph->vertices[i-1];
			graph->vertices[i].right = (i == width-1) ? NULL : &graph->vertices[i+1];
		}
		else if (i % width == 0) // Else if it is a left cell of the maze
		{
			graph->vertices[i].top = &graph->vertices[i-width];
			graph->vertices[i].bottom = (i == size - width) ? NULL : &graph->vertices[i+width];
			graph->vertices[i].left = NULL;
			graph->vertices[i].right = (width == 1) ? NULL : &graph->vertices[i+1];
		}
		else if (i % width == width-1) // Else if it is a right cell of the maze
		{
			graph->vertices[i].top = &graph->vertices[i-width];
			graph->vertices[i].bottom = (i == size - 1) ? NULL : &graph->vertices[i+width];
			graph->vertices[i].left = &graph->vertices[i-1];
			graph->vertices[i].right = NULL;
		}
		else if (i >= size - width) // Else if it is a bottom cell of the maze
		{
			graph->vertices[i].top = &graph->vertices[i-width];
			graph->vertices[i].bottom = NULL;
			graph->vertices[i].left = &graph->vertices[i-1];
			graph->vertices[i].right = &graph->vertices[i+1];
		}
		else // Else it is a cell inside of the maze (not an edge)
		{
			graph->vertices[i].top = &graph->vertices[i-width];
			graph->vertices[i].bottom = &graph->vertices[i+width];
			graph->vertices[i].left = &graph->vertices[i-1];
			graph->vertices[i].right = &graph->vertices[i+1];
		}
	}

	return graph;
}

void add_wall(struct graph *const graph, long long int const i, long long int const j)
{
	long long int const width = graph->width;
	long long int const height = graph->height;
	long long int const size = width * height;

	if (graph == NULL || graph->vertices == NULL || i < 0 || j < 0 || i >= size || j >= size)
		return;

	// If graph->vertices[j] is at the right of graph->vertices[i]
	// and graph->vertices[i] is not a right border cell
	if (i + 1 == j && i % width != width-1)
	{
		graph->vertices[i].right = NULL;
		graph->vertices[j].left = NULL;
	}
	// Else if graph->vertices[j] is at the bottom of graph->vertices[i]
	else if (i + width == j)
	{
		graph->vertices[i].bottom = NULL;
		graph->vertices[j].top = NULL;
	}
	// Else if graph->vertices[j] is at the left of graph->vertices[i]
	// and graph->vertices[i] is not a left border cell
	else if (i - 1 == j && i % width != 0)
	{
		graph->vertices[i].left = NULL;
		graph->vertices[j].right = NULL;
	}
	// Else if graph->vertices[j] is at the top of graph->vertices[i]
	else if (i - width == j)
	{
		graph->vertices[i].top = NULL;
		graph->vertices[j].bottom = NULL;
	}

	// Else graph->vertices[i] and graph->vertices[j] are not connected,
	// so no wall can separate them.
}

void generate_maze_aux(struct graph *const graph, long long int const first_vertex_id,
	long long int const cur_width, long long int const cur_height)
{
	if (cur_width == 1 || cur_height == 1)
		return;

	unsigned long long int const width = graph->width;

	if (cur_width >= cur_height)
	{
		long long int const r_vertex_id = random_int(first_vertex_id,
											first_vertex_id + cur_width - 2, 1);
		long long int const r_door_id = random_int(r_vertex_id,
											r_vertex_id + width * (cur_height - 1), width);
		long long int vertex_i;

		for (long long int i = 0; i < cur_height; i++)
		{
			vertex_i = r_vertex_id + width * i;

			if (vertex_i != r_door_id)
				add_wall(graph, vertex_i, vertex_i + 1);
		}

		long long int const left_width = r_vertex_id - first_vertex_id + 1;
		generate_maze_aux(graph, first_vertex_id, left_width, cur_height);
		generate_maze_aux(graph, r_vertex_id + 1, cur_width - left_width, cur_height);
	}
	else
	{
		long long int const r_vertex_id = random_int(first_vertex_id,
									width * (cur_height - 1 + first_vertex_id / width) - 1, width);
		long long int const r_door_id = random_int(r_vertex_id, r_vertex_id + cur_width - 1, 1);
		long long int vertex_i;

		for (long long int i = 0; i < cur_width; i++)
		{
			vertex_i = r_vertex_id + i;

			if (vertex_i != r_door_id)
				add_wall(graph, vertex_i, vertex_i + width);
		}

		long long int const top_height = (r_vertex_id - first_vertex_id) / width + 1;
		generate_maze_aux(graph, first_vertex_id, cur_width, top_height);
		generate_maze_aux(graph, r_vertex_id + width, cur_width, cur_height - top_height);
	}
}

void generate_maze(struct graph *const graph)
{
	generate_maze_aux(graph, 0, graph->width, graph->height);
}

void free_graph(struct graph *graph)
{
	if (graph != NULL)
	{
		if (graph->vertices != NULL)
			free(graph->vertices);

		free(graph);
	}
}

struct vertex **get_neighbours(struct vertex *const vertex)
{
	if (vertex == NULL)
		return NULL;

	struct vertex **neighbours = (struct vertex **) calloc(5, sizeof(struct vertex *));
	if (neighbours == NULL)
		return NULL;

	unsigned char i = 0;

	if (vertex->right != NULL && vertex->right->parent == NULL)
	{
		neighbours[i] = vertex->right;
		i++;
	}
	if (vertex->bottom != NULL && vertex->bottom->parent == NULL)
	{
		neighbours[i] = vertex->bottom;
		i++;
	}
	if (vertex->top != NULL && vertex->top->parent == NULL)
	{
		neighbours[i] = vertex->top;
		i++;
	}
	if (vertex->left != NULL && vertex->left->parent == NULL)
	{
		neighbours[i] = vertex->left;
		i++;
	}

	neighbours[i] = NULL;

	return neighbours;
}

struct vertex *find_exit(struct graph *const graph, long long int const start)
{
	long long int const size = graph->width * graph->height;

	if (graph == NULL || graph->vertices == NULL || start < 0 || start >= size)
		return NULL;

	struct vertex **queue = (struct vertex **) calloc(size, sizeof(struct vertex *));
	if (queue == NULL)
		return NULL;

	graph->vertices[start].parent = &graph->vertices[start];

	unsigned long long int queue_i = 0;
	unsigned long long int queue_size = 1;
	queue[queue_i] = &graph->vertices[start];


	unsigned char j = 0;
	struct vertex **neighbours;
	struct vertex *cur_neighbour;

	struct vertex *cur_vertex = queue[queue_i];
	
	while (!cur_vertex->is_end)
	{
		neighbours = get_neighbours(cur_vertex);
		if (neighbours == NULL)
		{
			free(queue);
			return NULL;
		}

		j = 0;
		cur_neighbour = neighbours[j];
		while (cur_neighbour != NULL)
		{
			cur_neighbour->parent = cur_vertex;
			if (cur_neighbour->is_end)
			{
				free(neighbours);
				free(queue);

				return cur_neighbour;
			}

			queue[queue_size++] = cur_neighbour;
			cur_neighbour = neighbours[++j];
		}

		cur_vertex = queue[++queue_i];
		free(neighbours);
	}

	free(queue);

	return cur_vertex;
}
