#include "queue.h"

/**
 * @brief Function to initialize the queue.
 * 
 * Initializes a queue with the given capacity. The queue is empty upon creation.
 * 
 * @param capacity The maximum number of elements the queue can hold.
 * @return A pointer to the newly created queue.
 */
queue_t *createQueue(int capacity) {
    queue_t *q = (queue_t *)malloc(sizeof(queue_t)); // Allocate memory for the queue structure.
    q->data = (int *)malloc(capacity * sizeof(int));  // Allocate memory for the queue's data array.
    q->front = 0;                                    // The front is at index 0 initially.
    q->size = 0;                                     // Queue is empty at initialization.
    q->capacity = capacity;                          // Set the queue's maximum capacity.
    q->rear = capacity - 1;                          // The rear starts at the last index.
    return q;
}

/**
 * @brief Function to add an element to the queue.
 * 
 * Adds an item to the rear of the queue. If the queue is full, it resizes to accommodate more elements.
 * 
 * @param q The queue to add the element to.
 * @param item The item to add to the queue.
 */
void enqueue(queue_t *q, int item) {
    if (q->size == q->capacity) {
        // Optionally, implement resizing logic here if needed.
        // This can involve reallocating memory for the queue's data array.
        return;
    }

    q->rear = (q->rear + 1) % q->capacity; // Use modulo to wrap around the rear if needed.
    q->data[q->rear] = item;                // Insert the item at the rear of the queue.
    q->size++;                               // Increase the size of the queue.
}

/**
 * @brief Function to remove an element from the queue.
 * 
 * Removes the element from the front of the queue and shifts the front pointer.
 * 
 * @param q The queue to remove the element from.
 * @return The item that was removed from the front of the queue.
 */
int dequeue(queue_t *q) {
    if (isEmpty(q)) {
        // Optionally, handle the case where the queue is empty.
        return -1; // Return a sentinel value indicating the queue is empty.
    }

    int item = q->data[q->front];             // Store the item at the front of the queue.
    q->front = (q->front + 1) % q->capacity;  // Use modulo to wrap around the front pointer.
    q->size--;                                // Decrease the size of the queue.
    return item;                              // Return the removed item.
}

/**
 * @brief Function to check if the queue is empty.
 * 
 * Checks if the queue has any elements by comparing the size to zero.
 * 
 * @param q The queue to check.
 * @return True if the queue is empty, false otherwise.
 */
bool isEmpty(queue_t *q) {
    return q->size == 0; // Queue is empty if size is zero.
}
