#include <stdlib.h>
#include <stdio.h>
#include <time.h>

struct vertex {
    int id;
    int is_start;
    int is_end;
    struct vertex *top;
    struct vertex *bottom;
    struct vertex *left;
    struct vertex *right;
};

int random_int(int min, int max, int const step)
{
	if (min > max)
	{
		int tmp = min;
		min = max;
		max = tmp;
	}

	int k = (max - min) / step;

	return min + step * (rand() % (k+1));
}

struct vertex *init_graph(int const width, int const height, int id_start)
{
	int const size = width * height;

	if (width < 1 || height < 1 || id_start < 0 || id_start >= size)
		return NULL;

	struct vertex *graph = (struct vertex *) calloc(size, sizeof(struct vertex));

	if (graph == NULL)
		return NULL;

	for (int i = 0; i < size; i++)
	{
		graph[i].id = i;
		graph[i].is_start = (i == id_start) ? 1 : 0;
		graph[i].is_end = (i == size-1) ? 1 : 0;

		if (i < width) // If it is a top cell of the maze
		{
			graph[i].top = NULL;
			graph[i].bottom = &graph[i+width];
			graph[i].left = (i == 0) ? NULL : &graph[i-1];
			graph[i].right = (i == width-1) ? NULL : &graph[i+1];
		}
		else if (i % width == 0) // Else if it is a left cell of the maze
		{
			graph[i].top = &graph[i-width];
			graph[i].bottom = (i == size - width) ? NULL : &graph[i+width];
			graph[i].left = NULL;
			graph[i].right = &graph[i+1];
		}
		else if (i % width == width-1) // Else if it is a right cell of the maze
		{
			graph[i].top = &graph[i-width];
			graph[i].bottom = (i == size - 1) ? NULL : &graph[i+width];
			graph[i].left = &graph[i-1];
			graph[i].right = NULL;
		}
		else if (i >= size - width) // Else if it is a bottom cell of the maze
		{
			graph[i].top = &graph[i-width];
			graph[i].bottom = NULL;
			graph[i].left = &graph[i-1];
			graph[i].right = &graph[i+1];
		}
		else // Else it is a cell inside of the maze (not an edge)
		{
			graph[i].top = &graph[i-width];
			graph[i].bottom = &graph[i+width];
			graph[i].left = &graph[i-1];
			graph[i].right = &graph[i+1];
		}
	}

	return graph;
}

void add_wall(struct vertex *const graph, int const i, int const j,
	int const width, int const size)
{
	if (graph == NULL || i < 0 || j < 0 || i >= size || j >= size)
		return;

	// If graph[j] is at the right of graph[i]
	// and graph[i] is not a right border cell
	if (i + 1 == j && i % width != width-1)
	{
		graph[i].right = NULL;
		graph[j].left = NULL;
	}
	// Else if graph[j] is at the bottom of graph[i]
	else if (i + width == j)
	{
		graph[i].bottom = NULL;
		graph[j].top = NULL;
	}
	// Else if graph[j] is at the left of graph[i]
	// and graph[i] is not a left border cell
	else if (i - 1 == j && i % width != 0)
	{
		graph[i].left = NULL;
		graph[j].right = NULL;
	}
	// Else if graph[j] is at the top of graph[i]
	else if (i - width == j)
	{
		graph[i].top = NULL;
		graph[j].bottom = NULL;
	}

	// Else graph[i] and graph[j] are not connected,
	// so no wall can separate them.
}

void generate_maze_aux(struct vertex *const graph, int const width, int const height,
	int const first_vertex_id, int const cur_width, int const cur_height)
{
	if (cur_width == 1 || cur_height == 1)
		return;

	int const size = width * height;

	if (cur_width >= cur_height)
	{
		int const r_vertex_id = random_int(first_vertex_id, first_vertex_id + cur_width - 2, 1);
		int const r_door_id = random_int(r_vertex_id, r_vertex_id + width * (cur_height - 1), width);
		int vertex_i;

		for (int i = 0; i < cur_height; i++)
		{
			vertex_i = r_vertex_id + width * i;

			if (vertex_i != r_door_id)
				add_wall(graph, vertex_i, vertex_i + 1, width, size);
		}

		int const left_width = r_vertex_id - first_vertex_id + 1;
		generate_maze_aux(graph, width, height, first_vertex_id, left_width, cur_height);
		generate_maze_aux(graph, width, height, r_vertex_id + 1, cur_width - left_width, cur_height);
	}
	else
	{
		int const r_vertex_id = random_int(first_vertex_id,
									width * (cur_height - 1 + first_vertex_id / width) - 1, width);
		int const r_door_id = random_int(r_vertex_id, r_vertex_id + cur_width - 1, 1);
		int vertex_i;

		for (int i = 0; i < cur_width; i++)
		{
			vertex_i = r_vertex_id + i;

			if (vertex_i != r_door_id)
				add_wall(graph, vertex_i, vertex_i + width, width, size);
		}

		int const top_height = (r_vertex_id - first_vertex_id) / width + 1;
		generate_maze_aux(graph, width, height, first_vertex_id, cur_width, top_height);
		generate_maze_aux(graph, width, height, r_vertex_id + width,
			cur_width, cur_height - top_height);
	}
}

void generate_maze(struct vertex *const graph, int const width, int const height)
{
	generate_maze_aux(graph, width, height, 0, width, height);
}

void free_graph(struct vertex *graph)
{
	if (graph != NULL)
		free(graph);
}

//----------------------- Test ------------------------------ */
int test_graph_connections(struct vertex *const graph, int const width, int const height)
{
    if (graph == NULL)
    {
        printf("Graph est nul\n");
        return 0;
    }

    int const size = width * height;

    for (int i = 0; i < size; i++)
    {
        struct vertex *v = &graph[i];

        // Vérification des voisins gauche/droite
        if (v->right)
        {
            if (v->right->left != v)
            {
                printf("Error: Vertex %d's right neighbor does not point back.\n", v->id);
                return 0;
            }
        }
        if (v->left)
        {
            if (v->left->right != v)
            {
                printf("Error: Vertex %d's left neighbor does not point back.\n", v->id);
                return 0;
            }
        }

        // Vérification des voisins haut/bas
        if (v->top)
        {
            if (v->top->bottom != v)
            {
                printf("Error: Vertex %d's top neighbor does not point back.\n", v->id);
                return 0;
            }
        }
        if (v->bottom)
        {
            if (v->bottom->top != v)
            {
                printf("Error: Vertex %d's bottom neighbor does not point back.\n", v->id);
                return 0;
            }
        }
    }

    // Si tout est correct, afficher les connexions de chaque vertex
    printf("Il n'y a pas d'erreur au niveaux de la connexions des voisin de chaque case voila\n");
    
    for (int i = 0; i < size; i++)
    {
        struct vertex *v = &graph[i];
        printf("\nVertex %d: \n", v->id+1);

        if (v->top)
            printf("Top: %d, \n", v->top->id+1);
        else
            printf("Top: NULL, \n");

        if (v->bottom)
            printf("Bottom: %d, \n", v->bottom->id+1);
        else
            printf("Bottom: NULL, \n");

        if (v->left)
            printf("Left: %d, \n", v->left->id+1);
        else
            printf("Left: NULL, \n");

        if (v->right)
            printf("Right: %d\n", v->right->id+1);
        else
            printf("Right: NULL\n");
    }

    return 1; // Succès
}


int main()
{
    // Dimensions du labyrinthe
    int width = 5;
    int height = 4;

    // Initialiser le générateur de nombres aléatoires
    srand(time(NULL));

    // Générer le graphe
    struct vertex *graph = init_graph(width, height, 0);

    if (graph == NULL)
    {
        printf("Erreur generation du graphe\n");
        return EXIT_FAILURE;
    }

    // Tester les connexions avant la génération du labyrinthe (sans murs)
    printf("Teste de la connexion des graphe avant le placement des murs\n");
    test_graph_connections(graph, width, height);

    // Générer le labyrinthe (ce qui place les murs)
    generate_maze(graph, width, height);

    printf("\nTest de la connexion des graphe après le placement des murs\n");
    test_graph_connections(graph, width, height);

    // Libérer la mémoire allouée au graphe
    free_graph(graph);

    return EXIT_SUCCESS;
}